package cn.dankal.share.common;

import cn.hutool.core.util.RandomUtil;

/**
 * int 数组工厂
 * <p>
 * 生成指定长度的 int 数组 并使用随机数填充
 * <br/>
 * 随机数范围: array.length*3
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class IntArrayFactory {

    /**
     * 生成的默认数组长度
     */
    private final static Integer DEFAULT_ARRAY_LENGTH = 1000000;

    /**
     * 生成默认长度的 int  数组
     *
     * @return int []
     * @see IntArrayFactory#DEFAULT_ARRAY_LENGTH
     */
    public static int[] generateIntArray() {
        return generateIntArray(DEFAULT_ARRAY_LENGTH);
    }

    /**
     * 生成指定长度的数组
     *
     * @param arrayLength 生成长度为 <code>arrayLength</code> int 数组
     * @return int []
     */
    public static int[] generateIntArray(int arrayLength) {
        if (arrayLength <= 0) {
            arrayLength = DEFAULT_ARRAY_LENGTH;
        }
        int[] resultArray = new int[arrayLength];
        for (int i = 0; i < resultArray.length; i++) {
            resultArray[i] = RandomUtil.randomInt(0, arrayLength * 3);
        }
        return resultArray;
    }
}
