package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * CountDownLatch 简单使用测试
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CountDownLatchTest {

    public static void main(String[] args) {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("countDownLatch-test");
        try {
        /*
          情景小析:
              我们需要做这样一件事情:
                  target: 获取商品详情
                    - 获取商品信息                 耗时 50 毫秒
                    - 获取商品价格                 耗时 10 毫秒
                    - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
                  - 假设三个子任务之间没有依赖关系
         */
            // target: 获取商品详细
            CountDownLatch countDownLatch = new CountDownLatch(3);

            final long startTime = System.currentTimeMillis();

            // 获取商品信息 50 毫秒
            final long getProductDetailElapsed = 50;
            Worker getProductDetailWorker = new Worker("获取商品信息", getProductDetailElapsed, countDownLatch);
            threadPoolExecutor.execute(getProductDetailWorker);

            // 获取商品价格 10 毫秒
            final long getPriceElapsed = 10;
            Worker getProductPrice = new Worker("获取商品价格", getPriceElapsed, countDownLatch);
            threadPoolExecutor.execute(getProductPrice);

            //商品属性获取并归并  65毫秒
            final long productPropertiesReduceElapsed = 65;
            Worker productPropertiesReduceWorker = new Worker("商品属性获取并归并", productPropertiesReduceElapsed, countDownLatch);
            threadPoolExecutor.execute(productPropertiesReduceWorker);

            try {
                System.out.printf("%s 线程即将开始等待 ... ...%n", Thread.currentThread().getName());
                countDownLatch.await();
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.println("接口返回~");
            System.out.println("接口放回理论耗时: " + (getProductDetailElapsed + getPriceElapsed + productPropertiesReduceElapsed) + " 毫秒");
            System.out.println("接口放回实际耗时: " + (System.currentTimeMillis() - startTime) + " 毫秒");
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}

