package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.HashMap;
import java.util.UUID;
import java.util.concurrent.Exchanger;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * {@link Exchanger} 的简单使用
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class ExchangerTest {

    public static void main(String[] args) {
        /*
             应用场景:
                 - 在两个线程之间进行数据交换

               - 流水线式数据处理
                   - 线程一 处理完数据之后,将数据交给 线程二继续处理
         */
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("exchanger-test");
        Exchanger<HashMap<String, Object>> mapExchanger = new Exchanger<>();
        try {
            // 线程一 先处理
            threadPoolExecutor.execute(() -> {
                final HashMap<String, Object> dataMap = new HashMap<>(1 << 5);
                dataMap.put(Thread.currentThread().getName(), UUID.randomUUID().toString().replace("-", ""));
                System.out.println(Thread.currentThread().getName() + " 开始处理数据");
                try {
                    TimeUnit.SECONDS.sleep(1);
                    System.out.printf("%s 线程 原始数据: %s%n", Thread.currentThread().getName(), dataMap);
                    System.out.println(Thread.currentThread().getName() + " 完成数据处理,即将进行数据传递");
                    HashMap<String, Object> exchangeResult = mapExchanger.exchange(dataMap);
                    System.out.printf("%s 获取交换后的结果 %s%n", Thread.currentThread().getName(), exchangeResult);
                } catch (InterruptedException e) {
                    // ignored
                }
            });
            // 线程二 继续处理
            threadPoolExecutor.execute(() -> {
                System.out.println(Thread.currentThread().getName() + " 开始处理数据");
                try {
                    HashMap<String, Object> emptyMap = new HashMap<>(1);
                    TimeUnit.SECONDS.sleep(1);
                    System.out.printf("%s 线程 原始数据: %s%n", Thread.currentThread().getName(), emptyMap);
                    System.out.println(Thread.currentThread().getName() + " 完成数据处理,即将进行数据传递");
                    HashMap<String, Object> exchangeResult = mapExchanger.exchange(emptyMap);
                    System.out.printf("%s 获取交换后的结果 %s%n", Thread.currentThread().getName(), exchangeResult);
                } catch (InterruptedException e) {
                    // ignored
                }
            });
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}
