package cn.dankal.share.concurrentutil;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.TimeUnit;

/**
 * {@link Runnable}
 * Worker 简单模拟
 * <p>
 * 用于测试 {@link CountDownLatch} 和 {@link CyclicBarrier} 的简单使用
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class Worker implements Runnable {

    /**
     * 工作内容
     */
    private final String workContent;

    /**
     * 工作消耗时间:单位 second
     */
    private final long workTimeInMilliSecond;

    /**
     * <code>CountDownLatch</code>
     */
    private CountDownLatch countDownLatch;

    /**
     * <code>CyclicBarrier</code>
     */
    private CyclicBarrier cyclicBarrier;

    public Worker(String workContent, long workTimeInMilliSecond, CountDownLatch countDownLatch) {
        this.workContent = workContent;
        this.workTimeInMilliSecond = workTimeInMilliSecond;
        this.countDownLatch = countDownLatch;
    }

    public Worker(String workContent, long workTimeInMilliSecond, CyclicBarrier cyclicBarrier) {
        this.workContent = workContent;
        this.workTimeInMilliSecond = workTimeInMilliSecond;
        this.cyclicBarrier = cyclicBarrier;
    }

    @Override
    public void run() {
        // 工作耗时
        try {
            TimeUnit.MILLISECONDS.sleep(workTimeInMilliSecond);
            System.out.printf("%s 正在: %s  预计耗时 %d 毫秒 %n", Thread.currentThread().getName(), workContent, workTimeInMilliSecond);
        } catch (InterruptedException e) {
            // ignored
        } finally {
            if (null != this.countDownLatch) {
                // 计数扣减
                this.countDownLatch.countDown();
            } else if (null != cyclicBarrier) {
                try {
                    // 等待全部执行完成
                    this.cyclicBarrier.await();
                } catch (Exception e) {
                    // ignored
                }
            }
        }
    }
}
