package cn.dankal.share.forkjoin;

import cn.dankal.share.common.InsertionSort;
import cn.dankal.share.common.MergeSort;

import java.util.Arrays;
import java.util.concurrent.RecursiveTask;

/**
 * 基于 {@link RecursiveTask} (ForkJoin) 简单实现 数组排序(归并排序)
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class IntArraySortTask extends RecursiveTask<int[]> {

    /**
     * 需要排序的数组
     */
    private final int[] intArray;

    /**
     * 数组拆分临界值(length)
     */
    private final int thresholdLength;

    public IntArraySortTask(int[] intArray, int thresholdLength) {
        this.intArray = intArray;
        this.thresholdLength = thresholdLength;
    }

    @Override
    protected int[] compute() {
        if (intArray.length <= thresholdLength) {
            // 达到拆分临界 直接排序
            return InsertionSort.doAscSort(intArray);
        } else {
            // formIndex ... middle ... toIndex
            // 任务拆分 将排序任务拆分成两个或更多

            // 此处演示 拆分为 左右两个任务
            int middle = intArray.length / 2;

            // 将目标数组从 “中间”拆分
            int[] leftIntArray = Arrays.copyOfRange(intArray, 0, middle);
            int[] rightIntArray = Arrays.copyOfRange(intArray, middle, intArray.length);
            // 拆分后,分为 左,右两个子Task
            IntArraySortTask leftSortTask = new IntArraySortTask(leftIntArray, thresholdLength);
            IntArraySortTask rightSortTask = new IntArraySortTask(rightIntArray, thresholdLength);
            //执行所有 Task
            invokeAll(leftSortTask, rightSortTask);
            // 归并 left  的结果
            int[] leftArray = leftSortTask.join();
            // 归并 right 的结果
            int[] rightArray = rightSortTask.join();
            // 合并 归并结果
            return MergeSort.mergeSortedArray(leftArray, rightArray);
        }
    }
}
