package cn.dankal.share.util.completablefuture;

import cn.dankal.share.common.ThreadExecutorFactory;
import cn.hutool.core.util.RandomUtil;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CompletableFutureAuxiliaryMethodTest {


    /**
     * ThreadPoolExecutor
     */
    private final static ThreadPoolExecutor THREAD_POOL_EXECUTOR = ThreadExecutorFactory.getThreadPoolExecutor("completableFuture-result-get-test");


    /**
     * 测试 CompletableFuture:allOf(CompletableFuture<?>... cfs)
     */
    @Test
    public void testAllOf() {
        Runnable runnable = () -> {
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("线程 %s 完成工作%n", Thread.currentThread().getName());
        };
        CompletableFuture<Void> completableFuture01 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture02 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture03 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture04 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture05 = CompletableFuture.runAsync(runnable);
        System.out.println("all completableFuture is starting the work or will start work");
        CompletableFuture.allOf(completableFuture01, completableFuture02, completableFuture03, completableFuture04, completableFuture05)
                .join();
        System.out.println("completableFuture all finished the work");
    }

    /**
     * 测试 CompletableFuture:anyOf(CompletableFuture<?>... cfs)
     */
    @Test
    public void testAnyOf() {
        Runnable runnable = () -> {
            try {
                TimeUnit.SECONDS.sleep(RandomUtil.randomInt(1,10));
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("线程 %s 完成工作%n", Thread.currentThread().getName());
        };
        CompletableFuture<Void> completableFuture01 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture02 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture03 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture04 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture05 = CompletableFuture.runAsync(runnable);
        System.out.println("all completableFuture is starting the work or will start work");
        CompletableFuture.anyOf(completableFuture01, completableFuture02, completableFuture03, completableFuture04, completableFuture05)
                .join();
        System.out.println("completableFuture all finished the work");
    }

    /**
     * 关闭 线程池资源
     */
    @AfterAll
    public static void after() {
        THREAD_POOL_EXECUTOR.shutdown();
    }
}
