package cn.dankal.share.util.completablefuture;

import cn.dankal.share.common.ThreadExecutorFactory;
import cn.hutool.core.util.RandomUtil;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CompletableFutureResultGetTest {

    /**
     * ThreadPoolExecutor
     */
    private final static ThreadPoolExecutor THREAD_POOL_EXECUTOR = ThreadExecutorFactory.getThreadPoolExecutor("completableFuture-result-get-test");

    /**
     * completableFuture 结果获取: get()
     */
    @Test
    public void testCompletableFutureGet() {
        // 使用 get() 获取 completableFuture 运行结果
        CompletableFuture<Integer> completableFuture = CompletableFuture.supplyAsync(() -> {
            int result = RandomUtil.randomInt(1, 100);
            System.out.printf("线程 %s 已生产随机数 %d%n", Thread.currentThread().getName(), result);
            return result;
        }, THREAD_POOL_EXECUTOR);
        try {
            System.out.printf("%s 线程 即将开始获取 completableFuture 结果%n", Thread.currentThread().getName());
            Integer result = completableFuture.get();
            System.out.println(result);
        } catch (Exception exception) {
            exception.printStackTrace();
        }
    }

    /**
     * completableFuture 结果获取: get(long timeout, TimeUnit unit)
     */
    @Test
    public void testCompletableFutureGetTimeout() {
        // 使用 get() 获取 completableFuture 运行结果
        CompletableFuture<Integer> completableFuture = CompletableFuture.supplyAsync(() -> {
            int result = RandomUtil.randomInt(1, 100);
            System.out.printf("线程 %s 已生产随机数 %d%n", Thread.currentThread().getName(), result);
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                // ignored
            }
            return result;
        }, THREAD_POOL_EXECUTOR);
        try {
            System.out.printf("%s 线程 即将开始获取 completableFuture 结果%n", Thread.currentThread().getName());
            Integer result = completableFuture.get(100, TimeUnit.MILLISECONDS);
            System.out.println(result);
        } catch (Exception exception) {
            System.out.println(exception.getMessage());
        }
    }

    /**
     * completableFuture 结果获取: getNow(T valueIfAbsent)
     */
    @Test
    public void testCompletableFutureGetNow() {
        // 使用 get() 获取 completableFuture 运行结果
        CompletableFuture<Integer> completableFuture = CompletableFuture.supplyAsync(() -> {
            int result = RandomUtil.randomInt(1, 100);
            System.out.printf("线程 %s 已生产随机数 %d%n", Thread.currentThread().getName(), result);
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                // ignored
            }
            return result;
        }, THREAD_POOL_EXECUTOR);
        try {
            System.out.printf("%s 线程 即将开始获取 completableFuture 结果%n", Thread.currentThread().getName());
            Integer result = completableFuture.getNow(-999);
            System.out.println(result);
        } catch (Exception exception) {
            System.out.println(exception.getMessage());
        }
    }

    /**
     * completableFuture 结果获取: join()
     */
    @Test
    public void testCompletableFutureJoin() {
        // 使用 get() 获取 completableFuture 运行结果
        CompletableFuture<Integer> completableFuture = CompletableFuture.supplyAsync(() -> {
            int result = RandomUtil.randomInt(1, 100);
            System.out.printf("线程 %s 已生产随机数 %d%n", Thread.currentThread().getName(), result);
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                // ignored
            }
            return result;
        }, THREAD_POOL_EXECUTOR);
        try {
            System.out.printf("%s 线程 即将开始获取 completableFuture 结果%n", Thread.currentThread().getName());
            Integer result = completableFuture.join();
            System.out.println(result);
        } catch (Exception exception) {
            System.out.println(exception.getMessage());
        }
    }

    /**
     * 关闭 线程池资源
     */
    @AfterAll
    public static void after() {
        THREAD_POOL_EXECUTOR.shutdown();
    }
}
