# 蛋壳创意科技-技术分享-Java常用并发工具类（讲稿版）

> company:  **[蛋壳创意科技](https://www.dankal.cn/)**
>
> code:  **[dankal-share-java-concurrent-util ](http://git.dankal.cn/mercyModest/dankal-share-java-concurrent-util.git)**

## 分享议题

> - **Fork/Join框架**
> - **Java常用并发工具类**
> - **组合式异步编程: CompletableFuture**

## `Fork-Join`

<b>		java 下多线程的开发可以我们自己启用多线程，线程池，还可以使用forkjoin，
forkjoin 可以让我们不去了解诸如Thread,Runnable 等相关的知识，只要遵循
forkjoin 的开发模式，就可以写出很好的多线程并发程序</b>

### 什么是分而治之

> ​		同时forkjoin 在处理某一类问题时非常的有用，哪一类问题？分而治之的问
> 题。
>
> 十大计算机经典算法
>
> - 快速排序
>
> - 堆排序
>
> - 归并排序
>
> - 二分查找
>
> - 线性查找、
>   深度优先
>
> - 广度优先
>
> - Dijkstra
>
> - 动态规划
>
> - 朴素贝叶斯分类
>
> 有几个属于分而治之？
>
> ​      3 个
>
> - 快速排序
>
> - 归并排序
>
> - 二分查找
>
> 分治法的设计思想是：将一个难以直接解决的大问题，分割成一些规模较小的相同问题，以便各个击破，分而治之。
> 
> ​    分治策略是：对于一个规模为n 的问题，若该问题可以容易地解决（比如说规模n 较小）则直接解决，否则将其分解为k 个规模较小的子问题，这些子问题互相独立且与原问题形式相同(子问题相互之间有联系就会变为动态规范算法)，递归地解这些子问题，然后将各子问题的解合并得到原问题的解。这种算法设计策略叫做分治法。

### `Fork-Join`原理

![](https://img.mercymodest.com/public/20210112142502.jpg)

#### 工作密取

​		即当前线程的Task 已经全被执行完毕，则自动取到其他线程的Task 池中取
出Task 继续执行。
​		ForkJoinPool 中维护着多个线程（一般为CPU 核数）在不断地执行Task，每
个线程除了执行自己职务内的Task 之外，还会根据自己工作线程的闲置情况去
获取其他繁忙的工作线程的Task，如此一来就能能够减少线程阻塞或是闲置的时
间，提高CPU 利用率。

![](https://img.mercymodest.com/public/20210112142618.jpg)

### `Fork-Join`编码实现

#### `Fork-Join`使用的标准范式

> ​		我们要使用ForkJoin 框架，必须首先创建一个ForkJoin 任务。它提供在任务
> 中执行fork 和join 的操作机制，通常我们不直接继承ForkjoinTask 类，只需要直
> 接继承其子类。
>
> 1. RecursiveAction，用于没有返回结果的任务
> 2. RecursiveTask，用于有返回值的任务
> task 要通过ForkJoinPool 来执行，使用submit 或invoke 提交，两者的区
> 别是：invoke 是同步执行，调用之后需要等待任务完成，才能执行后面的代码；
> submit 是异步执行。
> join()和get 方法当任务完成的时候返回计算结果。

![](https://img.mercymodest.com/public/20210112142903.jpg)

​		在我们自己实现的compute 方法里，首先需要判断任务是否足够小，如果足够小就直接执行任务。如果不足够小，就必须分割成两个子任务，每个子任务在调用invokeAll 方法时，又会进入compute 方法，看看当前子任务是否需要继续分割成孙任务，如果不需要继续分割，则执行当前子任务并返回结果。使用join方法会等待子任务执行完并得到其结果。

### `ForkJoin: RecursiveTask<V>`

**任务有返回值**

####  **使用ForkJoin简单实现归并排序.**

```java
package cn.dankal.share.forkjoin;

import cn.dankal.share.common.InsertionSort;
import cn.dankal.share.common.MergeSort;

import java.util.Arrays;
import java.util.concurrent.RecursiveTask;

/**
 * 基于 {@link RecursiveTask} (ForkJoin) 简单实现 数组排序(归并排序)
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class IntArraySortTask extends RecursiveTask<int[]> {

    /**
     * 需要排序的数组
     */
    private final int[] intArray;

    /**
     * 数组拆分临界值(length)
     */
    private final int thresholdLength;

    public IntArraySortTask(int[] intArray, int thresholdLength) {
        this.intArray = intArray;
        this.thresholdLength = thresholdLength;
    }

    @Override
    protected int[] compute() {
        if (intArray.length <= thresholdLength) {
            // 达到拆分临界 直接排序
            return InsertionSort.doAscSort(intArray);
        } else {
            // formIndex ... middle ... toIndex
            // 任务拆分 将排序任务拆分成两个或更多

            // 此处演示 拆分为 左右两个任务
            int middle = intArray.length / 2;

            // 将目标数组从 “中间”拆分
            int[] leftIntArray = Arrays.copyOfRange(intArray, 0, middle);
            int[] rightIntArray = Arrays.copyOfRange(intArray, middle, intArray.length);
            // 拆分后,分为 左,右两个子Task
            IntArraySortTask leftSortTask = new IntArraySortTask(leftIntArray, thresholdLength);
            IntArraySortTask rightSortTask = new IntArraySortTask(rightIntArray, thresholdLength);
            //执行所有 Task
            invokeAll(leftSortTask, rightSortTask);
            // 归并 left  的结果
            int[] leftArray = leftSortTask.join();
            // 归并 right 的结果
            int[] rightArray = rightSortTask.join();
            // 合并 归并结果
            return MergeSort.mergeSortedArray(leftArray, rightArray);
        }
    }
}
```

##### ForkJoin 性能评测: 归并排序实现

**对比对象：**

- **通常归并排序**

- **Arrays.sort（）**：DualPivotQuicksort 快速排序算法

**测试条件**

对 `3000000` int 数据 进行排序 

###### 单元测试

```java
    /**
     * 测试 归并排序对比
     */
    @Test
    public void testMergeSortThreeWays() {
        int[] generateIntArray = IntArrayFactory.generateIntArray(3000000);
        System.out.println("数组元素:  " + generateIntArray.length);

        // general merge sort
        final long generalMergeSortStart = System.currentTimeMillis();
        MergeSort.sort(generateIntArray);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - generalMergeSortStart) + " 毫秒 (general mergeSort)");

        // jdk utils: Arrays#sort
        final long arraySortStartMs = System.currentTimeMillis();
        // DualPivotQuicksort 快排
        Arrays.sort(generateIntArray);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - arraySortStartMs) + " 毫秒 (Arrays#sort)");

        // Fork/Join merge sort
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        IntArraySortTask intArraySortTask = new IntArraySortTask(generateIntArray, 300);
        final long forkJoinStartMs = System.currentTimeMillis();
        forkJoinPool.invoke(intArraySortTask);
        System.out.println("运行耗时:  " + (System.currentTimeMillis() - forkJoinStartMs) + " 毫秒 (Fork/Join mergeSort)");

    }
```

###### 运行结果

![](https://img.mercymodest.com/public/20210113115039.png)

**次要运行结果部分截图**

![](https://img.mercymodest.com/public/20210113144637.png)

**主要运行结果截图**

![](https://img.mercymodest.com/public/20210113144642.png)

### `ForkJoin:RecursiveAction`

**任务无返回值**

#### 使用 Fork/Join 框架实现目标文件搜索

##### `FindFileTask.java`

```java
package cn.dankal.share.forkjoin;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.io.FileUtil;
import cn.hutool.core.util.ArrayUtil;

import java.io.File;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.RecursiveAction;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * 基于 {@link RecursiveAction} (ForkJoin)实现 文件查找与简单统计
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class FindFileTask extends RecursiveAction {

    /**
     * 目标文件夹
     */
    private final String targetFolder;

    /**
     * 目标文件后缀
     */
    private final String findFileSuffix;

    /**
     * 目标文件计数器
     */
    public final static AtomicInteger COUNTER_TARGET_FILE = new AtomicInteger(0);

    /**
     * 遍历目录计数器
     */
    public final static AtomicInteger COUNTER_EACH_FOLDER = new AtomicInteger(0);


    public FindFileTask(String targetFolder, String findFileSuffix) {
        this.targetFolder = targetFolder;
        this.findFileSuffix = findFileSuffix;
    }

    @Override
    protected void compute() {
        // 参数校验
        if (!FileUtil.exist(targetFolder) || FileUtil.isFile(targetFolder)) {
            throw new IllegalArgumentException("targetFolder 必须是存在的 文件夹");
        }
        File targetFile = FileUtil.file(targetFolder);
        File[] files = targetFile.listFiles();
        if (ArrayUtil.isEmpty(files)) {
            // 空文件夹直接返回
            return;
        }
        // 需要执行 Task List
        List<FindFileTask> subFindFileTask = new ArrayList<>(1 << 5);
        for (File file : files) {
            if (file.isDirectory()) {
                // 文件夹
                FindFileTask findFileTask = new FindFileTask(file.getAbsolutePath(), findFileSuffix);
                subFindFileTask.add(findFileTask);
                COUNTER_EACH_FOLDER.incrementAndGet();
            } else {
                // 文件
                String fileSuffix = FileUtil.getSuffix(file);
                if (("." + fileSuffix).equalsIgnoreCase(findFileSuffix)) {
                    // 是目标文件
                    System.out.println("find: " + file.getName());
                    COUNTER_TARGET_FILE.incrementAndGet();
                }
            }
        }
        if (CollectionUtil.isNotEmpty(subFindFileTask)) {
            // 执行所有拆分的 task 如果需要的话
            for (FindFileTask findFileTask : invokeAll(subFindFileTask)) {
                // 因为是拆分 即使我们不需要获取拆分结果也需要 join 切切
                findFileTask.join();
            }
        }
    }
}
```

##### 单元测试

```java
   /**
     * 测试: Fork/Join RecursiveAction: 目标文件查找
     */
    @Test
    public void testFindFileByRecursiveAction() {
        final String targetFolder = "C:/dev/MavenRepository";
        final String findFileSuffix = ".lastUpdated";
        FindFileTask findFileTask = new FindFileTask(targetFolder, findFileSuffix);
        ForkJoinPool forkJoinPool = new ForkJoinPool();
        forkJoinPool.invoke(findFileTask);
        System.out.printf("共遍历 : %d 个文件夹%n", FindFileTask.COUNTER_EACH_FOLDER.get());
        System.out.printf("找到目标文件 %s : %d 个%n", findFileSuffix, FindFileTask.COUNTER_TARGET_FILE.get());
    }
```

##### 运行结果

![](https://img.mercymodest.com/public/20210113150306.png)

## 常用并发工具类

### Worker 模拟: 测试 `CountDownLatch`和`CyclicBarrier`

**Worker.java**

```java
package cn.dankal.share.concurrentutil;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.TimeUnit;

/**
 * {@link Runnable}
 * Worker 简单模拟
 * <p>
 * 用于测试 {@link CountDownLatch} 和 {@link CyclicBarrier} 的简单使用
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class Worker implements Runnable {

    /**
     * 工作内容
     */
    private final String workContent;

    /**
     * 工作消耗时间:单位 second
     */
    private final long workTimeInMilliSecond;

    /**
     * <code>CountDownLatch</code>
     */
    private CountDownLatch countDownLatch;

    /**
     * <code>CyclicBarrier</code>
     */
    private CyclicBarrier cyclicBarrier;

    public Worker(String workContent, long workTimeInMilliSecond, CountDownLatch countDownLatch) {
        this.workContent = workContent;
        this.workTimeInMilliSecond = workTimeInMilliSecond;
        this.countDownLatch = countDownLatch;
    }

    public Worker(String workContent, long workTimeInMilliSecond, CyclicBarrier cyclicBarrier) {
        this.workContent = workContent;
        this.workTimeInMilliSecond = workTimeInMilliSecond;
        this.cyclicBarrier = cyclicBarrier;
    }

    @Override
    public void run() {
        // 工作耗时
        try {
            TimeUnit.MILLISECONDS.sleep(workTimeInMilliSecond);
            System.out.printf("%s 正在: %s  预计耗时 %d 毫秒 %n", Thread.currentThread().getName(), workContent, workTimeInMilliSecond);
        } catch (InterruptedException e) {
            // ignored
        } finally {
            if (null != this.countDownLatch) {
                // 计数扣减
                this.countDownLatch.countDown();
            } else if (null != cyclicBarrier) {
                try {
                    // 等待全部执行完成
                    this.cyclicBarrier.await();
                } catch (Exception e) {
                    // ignored
                }
            }
        }
    }
}
```

#### CountDownLatch

##### 介绍

> ​		闭锁，CountDownLatch 这个类能够使一个线程等待其他线程完成各自的工作后再执行。例如，应用程序的主线程希望在负责启动框架服务的线程已经启动所有的框架服务之后再执行。
> ​		CountDownLatch 是通过一个计数器来实现的，计数器的初始值为初始任务的数量。每当完成了一个任务后，计数器的值就会减1（CountDownLatch.countDown()方法）。当计数器值到达0 时，它表示所有的已经完成了任务，然后在闭锁上等待CountDownLatch.await()方法的线程就可以恢复执行任务。
> ​	应用场景：
> ​		实现最大的并行性：有时我们想同时启动多个线程，实现最大程度的并行性。例如，我们想测试一个单例类。如果我们创建一个初始计数为1 的CountDownLatch，并让所有线程都在这个锁上等待，那么我们可以很轻松地完成测试。我们只需调用一次countDown()方法就可以让所有的等待线程同时恢复执行。
> ​		开始执行前等待n 个线程完成各自任务：例如应用程序启动类要确保在处理
> 用户请求前，所有N 个外部系统已经启动和运行了，例如处理excel 中多个表单。
>
> ![](https://img.mercymodest.com/public/20210113151952.png)

##### 模拟场景小析

> ```text
> 情景小析:
>     我们需要做这样一件事情:
>         target: 获取商品详情
>           - 获取商品信息                 耗时 50 毫秒
>           - 获取商品价格                 耗时 10 毫秒
>           - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
>         - 假设三个子任务之间没有依赖关系
> ```

##### `CountDownLatchTest.java`

```java
package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * CountDownLatch 简单使用测试
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CountDownLatchTest {

    public static void main(String[] args) {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("countDownLatch-test");
        try {
        /*
          情景小析:
              我们需要做这样一件事情:
                  target: 获取商品详情
                    - 获取商品信息                 耗时 50 毫秒
                    - 获取商品价格                 耗时 10 毫秒
                    - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
                  - 假设三个子任务之间没有依赖关系
         */
            // target: 获取商品详细
            CountDownLatch countDownLatch = new CountDownLatch(3);

            final long startTime = System.currentTimeMillis();

            // 获取商品信息 50 毫秒
            final long getProductDetailElapsed = 50;
            Worker getProductDetailWorker = new Worker("获取商品信息", getProductDetailElapsed, countDownLatch);
            threadPoolExecutor.execute(getProductDetailWorker);

            // 获取商品价格 10 毫秒
            final long getPriceElapsed = 10;
            Worker getProductPrice = new Worker("获取商品价格", getPriceElapsed, countDownLatch);
            threadPoolExecutor.execute(getProductPrice);

            //商品属性获取并归并  65毫秒
            final long productPropertiesReduceElapsed = 65;
            Worker productPropertiesReduceWorker = new Worker("商品属性获取并归并", productPropertiesReduceElapsed, countDownLatch);
            threadPoolExecutor.execute(productPropertiesReduceWorker);

            try {
                System.out.printf("%s 线程即将开始等待 ... ...%n", Thread.currentThread().getName());
                countDownLatch.await();
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.println("接口返回~");
            System.out.println("接口放回理论耗时: " + (getProductDetailElapsed + getPriceElapsed + productPropertiesReduceElapsed) + " 毫秒");
            System.out.println("接口放回实际耗时: " + (System.currentTimeMillis() - startTime) + " 毫秒");
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}
```

##### 运行结果

![](https://img.mercymodest.com/public/20210113150908.png)

#### CyclicBarrier

##### 介绍

> ​		CyclicBarrier 的字面意思是可循环使用（Cyclic）的屏障（Barrier）。它要做的事情是，让一组线程到达一个屏障（也可以叫同步点）时被阻塞，直到最后一个线程到达屏障时，屏障才会开门，所有被屏障拦截的线程才会继续运行。
> ​		CyclicBarrier 默认的构造方法是CyclicBarrier（int parties），其参数表示屏障拦截的线程数量，每个线程调用await 方法告诉CyclicBarrier 我已经到达了屏障，然后当前线程被阻塞。
> ​		CyclicBarrier 还提供一个更高级的构造函数CyclicBarrie（r int parties，Runnable
> barrierAction），用于在线程到达屏障时，优先执行barrierAction，方便处理更复杂的业务场景。
> ​		CyclicBarrier 可以用于多线程计算数据，最后合并计算结果的场景。
>
> ![](https://img.mercymodest.com/public/20210113152210.png)

##### 模拟场景小析

> ```text
> 情景小析:
>     我们需要做这样一件事情:
>         target: 获取商品详情
>           - 获取商品信息                 耗时 50 毫秒
>           - 获取商品价格                 耗时 10 毫秒
>           - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
>         - 假设三个子任务之间没有依赖关系
> ```

##### `CyclicBarrierTest.java`

```java
package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.concurrent.CyclicBarrier;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * CyclicBarrier 简单使用测试
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class CyclicBarrierTest {

    public static void main(String[] args) {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("countDownLatch-test");
        try {
        /*
          情景小析:
              我们需要做这样一件事情:
                  target: 获取商品详情
                    - 获取商品信息                 耗时 50 毫秒
                    - 获取商品价格                 耗时 10 毫秒
                    - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
                  - 假设三个子任务之间没有依赖关系
         */
            // target: 获取商品详细
            CyclicBarrier cyclicBarrier = new CyclicBarrier(4);

            final long startTime = System.currentTimeMillis();

            // 获取商品信息 50 毫秒
            final long getProductDetailElapsed = 50;
            Worker getProductDetailWorker = new Worker("获取商品信息", getProductDetailElapsed, cyclicBarrier);
            threadPoolExecutor.execute(getProductDetailWorker);

            // 获取商品价格 10 毫秒
            final long getPriceElapsed = 10;
            Worker getProductPrice = new Worker("获取商品价格", getPriceElapsed, cyclicBarrier);
            threadPoolExecutor.execute(getProductPrice);

            //商品属性获取并归并  65毫秒
            final long productPropertiesReduceElapsed = 65;
            Worker productPropertiesReduceWorker = new Worker("商品属性获取并归并", productPropertiesReduceElapsed, cyclicBarrier);
            threadPoolExecutor.execute(productPropertiesReduceWorker);

            try {
                System.out.printf("%s 线程即将开始等待 ... ...%n", Thread.currentThread().getName());
                cyclicBarrier.await();
            } catch (Exception e) {
                // ignored
            }
            System.out.println("接口返回~");
            System.out.println("接口放回理论耗时: " + (getProductDetailElapsed + getPriceElapsed + productPropertiesReduceElapsed) + " 毫秒");
            System.out.println("接口放回实际耗时: " + (System.currentTimeMillis() - startTime) + " 毫秒");
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}
```

##### 运行结果

![](https://img.mercymodest.com/public/20210113151050.png)

### CountDownLatch和CyclicBarrier

> ​		CountDownLatch 的计数器只能使用一次，而CyclicBarrier 的计数器可以反复
> 使用。
> ​		CountDownLatch.await 一般阻塞工作线程，所有的进行预备工作的线程执行
> countDown，而CyclicBarrier 通过工作线程调用await 从而自行阻塞，直到所有工
> 作线程达到指定屏障，再大家一起往下走。
> ​		在控制多个线程同时运行上，CountDownLatch 可以不限线程数量，而CyclicBarrier 是固定线程数。同时，CyclicBarrier 还可以提供一个barrierAction，合并多线程计算结果。

### Semaphore

#### 介绍

> ​		Semaphore（信号量）是用来控制同时访问特定资源的线程数量，它通过协调各个线程，以保证合理的使用公共资源。应用场景Semaphore 可以用于做流量控制，特别是公用资源有限的应用场景，比如数据库连接。假如有一个需求，要读取几万个文件的数据，因为都是IO 密集型任务，我们可以启动几十个线程并发读取，但是如果读到内存后，还需要存储到数据库中，而数据库的连接数只有10 个，这时我们必须控制只有10个线程同时获取数据库连接保存数据，否则会报错无法获取数据库连接。这个时候，就可以使用Semaphore 来做流量控制。Semaphore 的构造方法Semaphore（int permits）接受一个整型的数字，表示可用的许可证数量。Semaphore 的用法也很简单，首先线程使用Semaphore的acquire()方法获取一个许可证，使用完之后调用release()方法归还许可证。还可以用tryAcquire()方法尝试获取许可证。
>
> Semaphore 还提供一些其他方法，简略介绍如下。
>
> - intavailablePermits()：返回此信号量中当前可用的许可证数。
> - intgetQueueLength()：返回正在等待获取许可证的线程数。
> - booleanhasQueuedThreads()：是否有线程正在等待获取许可证。
> - void reducePermits（int reduction）：减少reduction 个许可证，是个protected方法。
> - Collection getQueuedThreads()：返回所有等待获取许可证的线程集合，是个protected 方法。
>
> ![](https://img.mercymodest.com/public/20210113152835.png)

#### 模拟场景小析

**使用 Semaphore 简单实现令牌桶限流**

#### `TokenBucket.java`

```java
package cn.dankal.share.concurrentutil;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Semaphore;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * token Bucket 简单模拟
 * <p>
 * 基于 {@link Semaphore} 简单实现
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class TokenBucket {

    /**
     * <code>Semaphore</code>
     */
    private final Semaphore semaphore;

    /**
     * <code>CountDownLatch</code>
     */
    private final CountDownLatch countDownLatch;

    /**
     * 令牌统计
     */
    private final AtomicInteger tokenCounter = new AtomicInteger(0);

    /**
     * 成功线程数统计
     */
    private final AtomicInteger successCounter = new AtomicInteger(0);

    /**
     * 失败线程数统计
     */
    private final AtomicInteger fairCounter = new AtomicInteger(0);

    public TokenBucket(Semaphore semaphore, CountDownLatch countDownLatch1) {
        this.semaphore = semaphore;
        this.countDownLatch = countDownLatch1;
    }

    /**
     * 获取令牌
     */
    public void getToken() {
        try {
            final long getTokenTimeout = 5;
            if (semaphore.tryAcquire(getTokenTimeout, TimeUnit.MILLISECONDS)) {
                // 成功获取到信号量: 获取令牌成功
                // 保证 tryAcquire 超时
                int tokenNum = tokenCounter.incrementAndGet();
                TimeUnit.SECONDS.sleep(1);
                System.out.printf("%s 成功获取到 %d 号令牌 %n", Thread.currentThread().getName(), tokenNum);
                successCounter.incrementAndGet();
            } else {
                // 获取令牌失败
                System.out.printf("%s 未获取到令牌 %n", Thread.currentThread().getName());
                fairCounter.incrementAndGet();
            }
        } catch (InterruptedException e) {
            // ignored
        } finally {
            // 成功/失败统计计数扣减
            this.countDownLatch.countDown();
        }

    }

    /**
     * 成功获取令桶统计
     *
     * @return 成功获取令牌统计
     */
    public int successCount() {
        return successCounter.get();
    }

    /**
     * 获取令牌失败统计
     *
     * @return 未成功获取令牌统计
     */
    public int fairCount() {
        return fairCounter.get();
    }
}
```

#### `SemaphoreTest.java`

```java
package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.concurrent.CountDownLatch;
import java.util.concurrent.Semaphore;
import java.util.concurrent.ThreadPoolExecutor;

/**
 * Semaphore 简单使用测试
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class SemaphoreTest {

    public static void main(String[] args) {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("semaphore-test");
        // semaphore: 令牌桶只有 3 个令牌
        Semaphore semaphore = new Semaphore(3);
        // 参与抢令牌的线程数
        final int count = 10;
        // 确保统计顺序执行完(成功/失败统计)
        CountDownLatch successAndFairCountDownLatch = new CountDownLatch(count);
        TokenBucket tokenBucket = new TokenBucket(semaphore, successAndFairCountDownLatch);
        try {
            for (int i = 0; i < count; i++) {
                threadPoolExecutor.execute(tokenBucket::getToken);
            }
            successAndFairCountDownLatch.await();
            System.out.println("参与线程数: " + count);
            System.out.println("成功线程数：" + tokenBucket.successCount());
            System.out.println("失败线程数：" + tokenBucket.fairCount());
        } catch (InterruptedException e) {
            // ignored
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}
```

#### 运行结果

![](https://img.mercymodest.com/public/20210113151337.png)

### Exchanger

#### 介绍

> ​		Exchanger（交换者）是一个用于线程间协作的工具类。Exchanger 用于进行线程间的数据交换。它提供一个同步点，在这个同步点，两个线程可以交换彼此的数据。这两个线程通过exchange 方法交换数据，如果第一个线程先执行exchange()方法，它会一直等待第二个线程也执行exchange 方法，当两个线程都到达同步点时，这两个线程就可以交换数据，将本线程生产出来的数据传递给对方。
>
> ![](https://img.mercymodest.com/public/20210113153358.png)
>
> 官方注释:
>
> ##### 原版：
>
> ```text
> 	A synchronization point at which threads can pair and swap elements within pairs. Each thread presents some object on entry to the exchange method, matches with a partner thread, and receives its partner's object on return. An Exchanger may be viewed as a bidirectional form of a SynchronousQueue. Exchangers may be useful in applications such as genetic algorithms and pipeline designs.
> 
>  Sample Usage: 
> 	Here are the highlights of a class that uses an Exchanger to swap buffers between threads so that the thread filling the buffer gets a freshly emptied one when it needs it, handing off the filled one to the thread emptying the buffer.
> ```
>
> ##### 大意可能是：
>
> ```text
> 	一个同步点，在这个点上，线程可以配对和交换配对内的元素。每个线程在进入交换方法时都会呈现一些对象，与伙伴线程进行匹配，并在返回时接收伙伴的对象。Exchanger可以看作是SynchronousQueue的一种双向形式。交换器在遗传算法和流水线设计等应用中可能很有用。
>  示例用法。
> 	下面是一个使用Exchanger在线程之间交换缓冲区的类的亮点，这样当需要时，填充缓冲区的线程就会得到一个新的空的缓冲区，把填充的缓冲区交给清空缓冲区的线程。
> ```
>
> ##### 官方示例
>
> ![](https://img.mercymodest.com/public/20210113153732.png)

#### `ExchangerTest.java`

```java
package cn.dankal.share.concurrentutil;

import cn.dankal.share.common.ThreadExecutorFactory;

import java.util.HashMap;
import java.util.UUID;
import java.util.concurrent.Exchanger;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

/**
 * {@link Exchanger} 的简单使用
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-02
 */
public class ExchangerTest {

    public static void main(String[] args) {
        /*
             应用场景:
                 - 在两个线程之间进行数据交换

               - 流水线式数据处理
                   - 线程一 处理完数据之后,将数据交给 线程二继续处理
         */
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("exchanger-test");
        Exchanger<HashMap<String, Object>> mapExchanger = new Exchanger<>();
        try {
            // 线程一 先处理
            threadPoolExecutor.execute(() -> {
                final HashMap<String, Object> dataMap = new HashMap<>(1 << 5);
                dataMap.put(Thread.currentThread().getName(), UUID.randomUUID().toString().replace("-", ""));
                System.out.println(Thread.currentThread().getName() + " 开始处理数据");
                try {
                    TimeUnit.SECONDS.sleep(1);
                    System.out.printf("%s 线程 原始数据: %s%n", Thread.currentThread().getName(), dataMap);
                    System.out.println(Thread.currentThread().getName() + " 完成数据处理,即将进行数据传递");
                    HashMap<String, Object> exchangeResult = mapExchanger.exchange(dataMap);
                    System.out.printf("%s 获取交换后的结果 %s%n", Thread.currentThread().getName(), exchangeResult);
                } catch (InterruptedException e) {
                    // ignored
                }
            });
            // 线程二 继续处理
            threadPoolExecutor.execute(() -> {
                System.out.println(Thread.currentThread().getName() + " 开始处理数据");
                try {
                    HashMap<String, Object> emptyMap = new HashMap<>(1);
                    TimeUnit.SECONDS.sleep(1);
                    System.out.printf("%s 线程 原始数据: %s%n", Thread.currentThread().getName(), emptyMap);
                    System.out.println(Thread.currentThread().getName() + " 完成数据处理,即将进行数据传递");
                    HashMap<String, Object> exchangeResult = mapExchanger.exchange(emptyMap);
                    System.out.printf("%s 获取交换后的结果 %s%n", Thread.currentThread().getName(), exchangeResult);
                } catch (InterruptedException e) {
                    // ignored
                }
            });
        } finally {
            threadPoolExecutor.shutdown();
        }
    }
}
```

## 组合式异步编程: `CompletableFuture`

![](https://img.mercymodest.com/public/20210113162009.png)

### `Future`的不足

> ​		Future 是Java 5 添加的类，用来描述一个异步计算的结果。你可以使用isDone方法检查计算是否完成，或者使用get 阻塞住调用线程，直到计算完成返回结果，你也可以使用cancel 方法停止任务的执行。虽然Future 以及相关使用方法提供了异步执行任务的能力，但是对于结果的获取却是很不方便，只能通过阻塞或者轮询的方式得到任务的结果。阻塞的方式显然和我们的异步编程的初衷相违背，轮询的方式又会耗费无谓的CPU 资源，而且也不能及时地得到计算结果，为什么不能用观察者设计模式当计算结果完成及时通知监听者呢？。
> ​		Java 的一些框架，比如Netty，自己扩展了Java 的Future 接口，提供了addListener 等多个扩展方法，Google guava 也提供了通用的扩展Future:ListenableFuture、SettableFuture 以及辅助类Futures 等,方便异步编程。
> ​		同时Future 接口很难直接表述多个Future 结果之间的依赖性。实际开发中，我们经常需要达成以下目的：
> ​		将两个异步计算合并为一个——这两个异步计算之间相互独立，同时第二个
> 又依赖于第一个的结果。
> ​		等待Future 集合中的所有任务都完成。
> ​		仅等待Future 集合中最快结束的任务完成（有可能因为它们试图通过不同
> 的方式计算同一个值），并返回它的结果。
> ​		应对Future 的完成事件（即当Future 的完成事件发生时会收到通知，并
> 能使用Future 计算的结果进行下一步的操作，不只是简单地阻塞等待操作的结
> 果）

### `CompletableFuture`

> ​		JDK1.8 才新加入的一个实现类CompletableFuture，实现了Future<T>，CompletionStage<T>两个接口。实现了Future 接口，意味着可以像以前一样通过阻塞或者轮询的方式获得结果

### CompletableFuture 的创建

**除了直接new 出一个CompletableFuture 的实例，还可以通过工厂方法创建CompletableFuture 的实例**

#### 工厂方法

![](https://img.mercymodest.com/public/20210113162455.png)

> ​		Asynsc 表示异步,而supplyAsync 与runAsync 不同在与前者异步返回一个结果,后者是void.第二个函数第二个参数表示是用我们自己创建的线程池,否则采用默认的ForkJoinPool.commonPool()作为它的线程池。

### 获取CompletableFuture结果集

> - public T get()
>
> - public T get(long timeout, TimeUnit unit)
>
> - public T getNow(T valueIfAbsent)
>
> - public T join()
>
> ​    getNow 有点特殊，如果结果已经计算完则返回结果或者抛出异常，否则返回给定的valueIfAbsent 值。
> ​    join 返回计算的结果或者抛出一个unchecked 异常(CompletionException)，它和get 对抛出的异常的处理有些细微的区别。

#### `get()`

![](https://img.mercymodest.com/public/20210113164218.png)

#### `get(long timeout, TimeUnit unit)`

![](https://img.mercymodest.com/public/20210113164238.png)

#### `getNow(T valueIfAbsent)`

![](https://img.mercymodest.com/public/20210113165130.png)

#### `join()`

![](https://img.mercymodest.com/public/20210113165414.png)

### CompletableFuture辅助方法

#### `allOf(CompletableFuture<?>... cfs)`

![](https://img.mercymodest.com/public/20210113165913.png)

##### 单元测试

```java
    /**
     * 测试 CompletableFuture:allOf(CompletableFuture<?>... cfs)
     */
    @Test
    public void testAllOf() {
        Runnable runnable = () -> {
            try {
                TimeUnit.SECONDS.sleep(1);
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("线程 %s 完成工作%n", Thread.currentThread().getName());
        };
        CompletableFuture<Void> completableFuture01 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture02 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture03 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture04 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture05 = CompletableFuture.runAsync(runnable);
        System.out.println("all completableFuture is starting the work or will start work");
        CompletableFuture.allOf(completableFuture01, completableFuture02, completableFuture03, completableFuture04, completableFuture05)
                .join();
        System.out.println("completableFuture all finished the work");
    }
```

##### 运行结果

![](https://img.mercymodest.com/public/20210113171254.png)

#### `anyOf(CompletableFuture<?>... cfs)`

![](https://img.mercymodest.com/public/20210113171422.png)

##### 单元测试

```java
    /**
     * 测试 CompletableFuture:anyOf(CompletableFuture<?>... cfs)
     */
    @Test
    public void testAnyOf() {
        Runnable runnable = () -> {
            try {
                TimeUnit.SECONDS.sleep(RandomUtil.randomInt(1,10));
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("线程 %s 完成工作%n", Thread.currentThread().getName());
        };
        CompletableFuture<Void> completableFuture01 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture02 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture03 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture04 = CompletableFuture.runAsync(runnable);
        CompletableFuture<Void> completableFuture05 = CompletableFuture.runAsync(runnable);
        System.out.println("all completableFuture is starting the work or will start work");
        CompletableFuture.anyOf(completableFuture01, completableFuture02, completableFuture03, completableFuture04, completableFuture05)
                .join();
        System.out.println("completableFuture all finished the work");
    }
```

##### 运行结果

![](https://img.mercymodest.com/public/20210113171657.png)

### CompletableFuture 测试用例

前面我们在测试 `CountdownLatch`和`Cyclicbarrier`的时候有如下需求

> ```text
>                 - 获取商品信息                 耗时 50 毫秒
>                 - 获取商品价格                 耗时 10 毫秒
>                 - 商品属性获取并归并(Sku/Spu)   耗时 65 毫秒
> ```

我们使用使用 `CountdownLatch` 和 `Cyclicbarrier` 实现上述需求的时候是假设 三者之间没有任何依赖关系.如果

上述需求变更为

> ```text
>             - 获取商品信息   (A)              耗时 50 毫秒
>             - 获取商品价格     (B)            耗时 10 毫秒
>             - 商品属性获取并归并(Sku/Spu)  (C)  耗时 65 毫秒 
>     重要: 我们只有获取到了 '商品信息' 才能执行 '商品属性获取并归并',即 C 必须在 A 之后执行      
> ```

显然 这个需求无法使用 `CountdownLatch` 和 `Cyclicbarrier`实现，但是我们借助  `CompletableFuture`API可以轻而易举实现.

#### 实现需求

>             - 获取商品信息   (A)              耗时 120 毫秒
>             - 获取商品价格     (B)            耗时 95 毫秒
>             - 商品属性获取并归并(Sku/Spu)  (C)  耗时 165 毫秒 
>     重要: 我们只有获取到了 '商品信息' 才能执行 '商品属性获取并归并',即 C 必须在 A 之后执行   

#### 编码实现

```java
package cn.dankal.share.util.completablefuture;

import cn.dankal.share.common.ThreadExecutorFactory;
import org.junit.jupiter.api.Test;

import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;
import java.util.function.Supplier;

/**
 * <p>
 * <code>CompletableFuture</code> 测试
 * </p>
 *
 * @author ZGH.MercyModest
 * @version V1.0.0
 * @create 2021-01-22
 */
public class CompletableFutureTest {

    /*
         需求:
              - 获取商品信息   (A)              耗时 120 毫秒
              - 获取商品价格     (B)            耗时 95 毫秒
              - 商品属性获取并归并(Sku/Spu)  (C)  耗时 165 毫秒
          重要: 我们只有获取到了 '商品信息' 才能执行 '商品属性获取并归并',即 C 必须在 A 之后执行
     */

    /**
     * <code>CompletableFuture</code> 测试
     */
    @Test
    public void test() throws Exception {
        ThreadPoolExecutor threadPoolExecutor = ThreadExecutorFactory.getThreadPoolExecutor("dankal-test-completableFuture");

        // 获取商品信息 耗时:  120 毫秒
        final long getProductElapsed = 120L;
        Supplier<Integer> getProductSupplier = () -> {
            try {
                // do work
                TimeUnit.MILLISECONDS.sleep(getProductElapsed);
            } catch (InterruptedException e) {
                // ignored
            }
            // 返回商品信息成功: 200
            System.out.printf("%s 获取商品信息成功,即将返回%n", Thread.currentThread().getName());
            return 200;
        };

        // 获取商品价格 耗时 95 毫秒
        final long getProductPriceElapsed = 95L;
        Supplier<Double> getProductPriceSupplier = () -> {
            try {
                // do work
                TimeUnit.MILLISECONDS.sleep(getProductPriceElapsed);
            } catch (InterruptedException e) {
                // ignored
            }
            System.out.printf("%s 成功获取到商品价格即将开始返回%n", Thread.currentThread().getName());
            return 135.53;
        };

        // 归并商品属性耗时  165 毫秒
        final long getProductPropertiesElapsed = 165L;
        //程序开始计时
        final long startMs = System.currentTimeMillis();
        // 获取 商品信息 并归并商品属性
        CompletableFuture<Integer> productAboutCompletableFuture = CompletableFuture.supplyAsync(getProductSupplier, threadPoolExecutor)
                .thenApplyAsync(previousResult -> {
                    System.out.printf("%s 获取到商品信息: %d 工耗时 %d 毫秒%n", Thread.currentThread().getName(), previousResult, (System.currentTimeMillis() - startMs));
                    try {
                        // do work
                        TimeUnit.MILLISECONDS.sleep(getProductPropertiesElapsed);
                    } catch (InterruptedException e) {
                        // ignored
                    }
                    System.out.printf("%s 完成商品属性归并,即将返回结果%n", Thread.currentThread().getName());
                    return previousResult * 2;
                }, threadPoolExecutor);
        CompletableFuture<Double> getProductPriceCompletableFuture = CompletableFuture.supplyAsync(getProductPriceSupplier, threadPoolExecutor);

        // 等待程序执行完成
        CompletableFuture.allOf(productAboutCompletableFuture, getProductPriceCompletableFuture).get();
        System.out.println("程序执行完成:");
        System.out.printf("\t运行耗时: %d %n", (System.currentTimeMillis() - startMs));
        System.out.printf("\t理论耗时: %d %n", (getProductElapsed + getProductPropertiesElapsed + getProductPriceElapsed));
        // 模拟 Service 方法返回
        System.out.printf("获取商品详情: %d%n", productAboutCompletableFuture.get()/2);
        System.out.printf("获取商品属性: %d%n", productAboutCompletableFuture.get());
        System.out.printf("获取商品价格: %.2f RMB%n", getProductPriceCompletableFuture.get());
    }
}
```

#### 运行结果

![](https://img.mercymodest.com/public/20210122165021.png)